"""Generate Factur-X invoices.

Currently, only the EN 16931/COMFORT profile is supported.
"""

from __future__ import annotations

import datetime
import xml.etree.ElementTree as ET
from base64 import b64encode

from .const import NS_CII, NS_RAM, NS_UDT
from .model import (
    BasicInvoice,
    BasicWLInvoice,
    DocumentAllowance,
    DocumentCharge,
    EN16931Invoice,
    EN16931LineItem,
    IncludedNote,
    LineAllowance,
    LineCharge,
    LineItem,
    MinimumInvoice,
    PaymentMeans,
    PaymentTerms,
    PostalAddress,
    ProductCharacteristic,
    ProductClassification,
    ReferenceDocument,
    Tax,
    TradeContact,
    TradeParty,
)
from .money import Money
from .type_codes import DocumentTypeCode
from .types import ID, DocRef, OptionalQuantity, Quantity

__all__ = [
    "generate_et",
    "generate_xml",
]

#
# XML Utility Functions
#


def xml_date(date: datetime.date) -> str:
    """Format a date for Factur-X XML."""
    return "{:04d}{:02d}{:02d}".format(date.year, date.month, date.day)


#
# Common Elements
#


def _date_element(
    parent: ET.Element, name: str, date: datetime.date
) -> ET.Element:
    el = ET.SubElement(parent, name)
    date_el = ET.SubElement(el, "udt:DateTimeString", format="102")
    date_el.text = "{:04d}{:02d}{:02d}".format(date.year, date.month, date.day)
    return el


def _id_element(parent: ET.Element, name: str, id: str) -> ET.Element:
    el = ET.SubElement(parent, name)
    ET.SubElement(el, "ram:ID").text = id
    return el


def _scheme_id_element(parent: ET.Element, name: str, id: ID) -> ET.Element:
    attrs: dict[str, str] = {}
    if id[1] is not None:
        attrs["schemeID"] = id[1]
    el = ET.SubElement(parent, name, attrs)
    el.text = id[0]
    return el


def _currency_element(
    parent: ET.Element,
    name: str,
    amount: Money,
    default_currency: str,
    *,
    with_currency: bool = False,
) -> ET.Element:
    el = ET.SubElement(parent, name)
    if with_currency or amount.currency != default_currency:
        el.set("currencyID", amount.currency)
    el.text = str(amount.amount)
    return el


def _quantity_element(
    parent: ET.Element, name: str, quantity: Quantity | OptionalQuantity
) -> ET.Element:
    q, unit = quantity
    attribs: dict[str, str] = {}
    if unit is not None:
        attribs["unitCode"] = unit
    el = ET.SubElement(parent, name, attribs)
    el.text = str(q)
    return el


def _email_element(
    parent: ET.Element, name: str, email_address: str
) -> ET.Element:
    el = ET.SubElement(parent, name)
    sub = ET.SubElement(el, "ram:URIID", schemeID="EM")
    sub.text = f"mailto:{email_address}"
    return el


def _address_element(parent: ET.Element, address: PostalAddress) -> None:
    root = ET.SubElement(parent, "ram:PostalTradeAddress")
    if address.post_code is not None:
        ET.SubElement(root, "ram:PostcodeCode").text = address.post_code
    if address.line_one is not None:
        ET.SubElement(root, "ram:LineOne").text = address.line_one
    if address.line_two is not None:
        ET.SubElement(root, "ram:LineTwo").text = address.line_two
    if address.line_three is not None:
        ET.SubElement(root, "ram:LineThree").text = address.line_three
    if address.city is not None:
        ET.SubElement(root, "ram:CityName").text = address.city
    ET.SubElement(root, "ram:CountryID").text = address.country_code
    if address.country_subdivision is not None:
        ET.SubElement(
            root, "ram:CountrySubDivisionName"
        ).text = address.country_subdivision


def _note_element(parent: ET.Element, note: IncludedNote) -> None:
    note_el = ET.SubElement(parent, "ram:IncludedNote")
    ET.SubElement(note_el, "ram:Content").text = note.content
    if note.subject_code is not None:
        ET.SubElement(note_el, "ram:SubjectCode").text = note.subject_code


def _document_element(parent: ET.Element, name: str, id: str | None) -> None:
    if id is None:
        return
    el = ET.SubElement(parent, name)
    ET.SubElement(el, "ram:IssuerAssignedID").text = id


def _generate_trade_party(
    parent: ET.Element, name: str, party: TradeParty
) -> None:
    el = ET.SubElement(parent, name)
    for id in party.ids:
        ET.SubElement(el, "ram:ID").text = id
    for global_id in party.global_ids:
        _scheme_id_element(el, "ram:GlobalID", global_id)
    if party.name is not None:
        ET.SubElement(el, "ram:Name").text = party.name
    if party.description is not None:
        ET.SubElement(el, "ram:Description").text = party.description
    if party.legal_id is not None or party.trading_business_name is not None:
        legal_el = ET.SubElement(el, "ram:SpecifiedLegalOrganization")
        if party.legal_id is not None:
            _scheme_id_element(legal_el, "ram:ID", party.legal_id)
        if party.trading_business_name is not None:
            ET.SubElement(
                legal_el, "ram:TradingBusinessName"
            ).text = party.trading_business_name
    if party.contact is not None:
        _generate_trade_contact(el, party.contact)
    if party.address is not None:
        _address_element(el, party.address)
    if party.email is not None:
        _email_element(el, "ram:URIUniversalCommunication", party.email)
    if party.tax_number is not None:
        tax = ET.SubElement(el, "ram:SpecifiedTaxRegistration")
        ET.SubElement(tax, "ram:ID", schemeID="FC").text = party.tax_number
    if party.vat_id is not None:
        tax = ET.SubElement(el, "ram:SpecifiedTaxRegistration")
        ET.SubElement(tax, "ram:ID", schemeID="VA").text = party.vat_id


def _generate_trade_contact(parent: ET.Element, contact: TradeContact) -> None:
    el = ET.SubElement(parent, "ram:DefinedTradeContact")
    if contact.person_name is not None:
        ET.SubElement(el, "ram:PersonName").text = contact.person_name
    if contact.department_name is not None:
        ET.SubElement(el, "ram:DepartmentName").text = contact.department_name
    if contact.phone is not None:
        phone_el = ET.SubElement(el, "ram:TelephoneUniversalCommunication")
        ET.SubElement(phone_el, "ram:CompleteNumber").text = contact.phone
    if contact.email is not None:
        _email_element(el, "ram:EmailURIUniversalCommunication", contact.email)


#
# XML Generation
#


def generate_et(invoice: MinimumInvoice) -> ET.Element:
    """
    Generate a Factur-X invoice as ElementTree.

    >>> from datetime import date
    >>> from decimal import Decimal
    >>> from pycheval.type_codes import DocumentTypeCode
    >>> from pycheval.model import MinimumInvoice, PostalAddress, TradeParty
    >>> invoice = MinimumInvoice(
    ...         "12345",
    ...         DocumentTypeCode.INVOICE,
    ...         date(2024, 8, 22),
    ...         TradeParty("Seller GmbH", PostalAddress("DE")),
    ...         TradeParty("Buyer AG", PostalAddress("DE")),
    ...         "EUR",
    ...         tax_basis_total_amount=(Decimal("10000.00"), "EUR"),
    ...         tax_total_amount=(Decimal("1900.00"), "EUR"),
    ...         grand_total_amount=(Decimal("11900.00"), "EUR"),
    ...         due_payable_amount=(Decimal("11900.00"), "EUR"),
    ... )
    >>> root = generate_et(invoice)
    """

    root = ET.Element(
        "rsm:CrossIndustryInvoice",
        {
            "xmlns:rsm": NS_CII,
            "xmlns:ram": NS_RAM,
            "xmlns:udt": NS_UDT,
        },
    )

    _generate_doc_context(root, invoice)
    _generate_doc(root, invoice)
    _generate_transaction(root, invoice)

    return root


def generate_xml(invoice: MinimumInvoice) -> str:
    """
    Generate a Factur-X invoice as XML string.

    >>> from datetime import date
    >>> from decimal import Decimal
    >>> from pycheval.type_codes import DocumentTypeCode
    >>> from pycheval.model import MinimumInvoice, PostalAddress, TradeParty
    >>> invoice = MinimumInvoice(
    ...         "12345",
    ...         DocumentTypeCode.INVOICE,
    ...         date(2024, 8, 22),
    ...         TradeParty("Seller GmbH", PostalAddress("DE")),
    ...         TradeParty("Buyer AG", PostalAddress("DE")),
    ...         "EUR",
    ...         tax_basis_total_amount=(Decimal("10000.00"), "EUR"),
    ...         tax_total_amount=(Decimal("1900.00"), "EUR"),
    ...         grand_total_amount=(Decimal("11900.00"), "EUR"),
    ...         due_payable_amount=(Decimal("11900.00"), "EUR"),
    ... )
    >>> xml_string = generate(invoice)
    """

    root = generate_et(invoice)
    return ET.tostring(root, encoding="unicode", xml_declaration=True)


def _generate_doc_context(parent: ET.Element, invoice: MinimumInvoice) -> None:
    doc_ctx = ET.SubElement(parent, "rsm:ExchangedDocumentContext")
    if invoice.business_process_id is not None:
        business_el = ET.SubElement(
            doc_ctx, "ram:BusinessProcessSpecifiedDocumentContextParameter"
        )
        ET.SubElement(business_el, "ram:ID").text = invoice.business_process_id
    # Specify the used profile.
    guideline_el = ET.SubElement(
        doc_ctx, "ram:GuidelineSpecifiedDocumentContextParameter"
    )
    ET.SubElement(guideline_el, "ram:ID").text = invoice.PROFILE_URN


def _generate_doc(parent: ET.Element, invoice: MinimumInvoice) -> None:
    doc = ET.SubElement(parent, "rsm:ExchangedDocument")
    ET.SubElement(doc, "ram:ID").text = invoice.invoice_number
    ET.SubElement(doc, "ram:TypeCode").text = str(invoice.type_code)
    _date_element(doc, "ram:IssueDateTime", invoice.invoice_date)
    if isinstance(invoice, BasicWLInvoice):
        for note in invoice.notes:
            _note_element(doc, note)


def _generate_transaction(parent: ET.Element, invoice: MinimumInvoice) -> None:
    transaction_el = ET.SubElement(parent, "rsm:SupplyChainTradeTransaction")
    if isinstance(invoice, BasicInvoice):
        assert len(invoice.line_items) >= 1  # BG-25
        for li in invoice.line_items:
            _generate_line_item(transaction_el, invoice, li)
    _generate_trade_agreement(transaction_el, invoice)
    _generate_delivery(transaction_el, invoice)
    _generate_settlement(transaction_el, invoice)


def _generate_line_item(
    parent: ET.Element, invoice: BasicInvoice, line_item: LineItem
) -> None:
    li_el = ET.SubElement(
        parent,
        "ram:IncludedSupplyChainTradeLineItem",
    )
    _generate_line_item_doc(li_el, line_item)
    _generate_line_item_product(li_el, line_item)
    _generate_line_trade_agreement(li_el, invoice, line_item)
    _generate_line_delivery(li_el, line_item)
    _generate_line_settlement(li_el, invoice, line_item)


def _generate_line_item_doc(parent: ET.Element, line_item: LineItem) -> None:
    li_doc = ET.SubElement(parent, "ram:AssociatedDocumentLineDocument")
    ET.SubElement(li_doc, "ram:LineID").text = line_item.id
    if isinstance(line_item, EN16931LineItem) and line_item.note is not None:
        _note_element(li_doc, line_item.note)


def _generate_line_item_product(
    parent: ET.Element, line_item: LineItem
) -> None:
    el = ET.SubElement(parent, "ram:SpecifiedTradeProduct")
    if line_item.global_id is not None:
        _scheme_id_element(el, "ram:GlobalID", line_item.global_id)
    if isinstance(line_item, EN16931LineItem):
        if line_item.seller_assigned_id is not None:
            ET.SubElement(
                el, "ram:SellerAssignedID"
            ).text = line_item.seller_assigned_id
        if line_item.buyer_assigned_id is not None:
            ET.SubElement(
                el, "ram:BuyerAssignedID"
            ).text = line_item.buyer_assigned_id
    ET.SubElement(el, "ram:Name").text = line_item.name
    if isinstance(line_item, EN16931LineItem):
        if line_item.description is not None:
            ET.SubElement(el, "ram:Description").text = line_item.description
        for characteristic in line_item.product_characteristics:
            _generate_product_characteristic(el, characteristic)
        for classification in line_item.product_classifications:
            _generate_product_classification(el, classification)
        if line_item.origin_country is not None:
            _id_element(el, "ram:OriginTradeCountry", line_item.origin_country)


def _generate_product_characteristic(
    parent: ET.Element, characteristic: ProductCharacteristic
) -> None:
    c_el = ET.SubElement(parent, "ram:ApplicableProductCharacteristic")
    ET.SubElement(c_el, "ram:Description").text = characteristic.description
    ET.SubElement(c_el, "ram:Value").text = characteristic.value


def _generate_product_classification(
    parent: ET.Element, classification: ProductClassification
) -> None:
    cl_el = ET.SubElement(parent, "ram:DesignatedProductClassification")
    cc_el = ET.SubElement(cl_el, "ram:ClassCode")
    cc_el.text = classification.class_code
    cc_el.attrib["listID"] = classification.list_id
    if classification.list_version_id is not None:
        cc_el.attrib["listVersionID"] = classification.list_version_id


def _generate_line_trade_agreement(
    parent: ET.Element, invoice: BasicInvoice, line_item: LineItem
) -> None:
    agreement = ET.SubElement(parent, "ram:SpecifiedLineTradeAgreement")
    if isinstance(line_item, EN16931LineItem):
        if line_item.buyer_order_line_id is not None:
            doc_el = ET.SubElement(
                agreement, "ram:BuyerOrderReferencedDocument"
            )
            ET.SubElement(
                doc_el, "ram:LineID"
            ).text = line_item.buyer_order_line_id
        if line_item.gross_unit_price is not None:
            price, quantity = line_item.gross_unit_price
            price_el = ET.SubElement(
                agreement, "ram:GrossPriceProductTradePrice"
            )
            _currency_element(
                price_el, "ram:ChargeAmount", price, invoice.currency_code
            )
            if quantity is not None:
                _quantity_element(price_el, "ram:BasisQuantity", quantity)
            if isinstance(line_item.gross_allowance_or_charge, LineAllowance):
                _generate_allowance_or_charge(
                    price_el,
                    "ram:AppliedTradeAllowanceCharge",
                    invoice,
                    line_item.gross_allowance_or_charge,
                    False,
                )
            elif isinstance(line_item.gross_allowance_or_charge, LineCharge):
                _generate_allowance_or_charge(
                    price_el,
                    "ram:AppliedTradeAllowanceCharge",
                    invoice,
                    line_item.gross_allowance_or_charge,
                    True,
                )
    price_el = ET.SubElement(agreement, "ram:NetPriceProductTradePrice")
    _currency_element(
        price_el,
        "ram:ChargeAmount",
        line_item.net_price,
        invoice.currency_code,
    )
    if line_item.basis_quantity is not None:
        _quantity_element(
            price_el, "ram:BasisQuantity", line_item.basis_quantity
        )


def _generate_line_delivery(parent: ET.Element, line_item: LineItem) -> None:
    delivery_el = ET.SubElement(parent, "ram:SpecifiedLineTradeDelivery")
    _quantity_element(
        delivery_el, "ram:BilledQuantity", line_item.billed_quantity
    )


def _generate_line_settlement(
    parent: ET.Element, invoice: BasicInvoice, line_item: LineItem
) -> None:
    settlement = ET.SubElement(parent, "ram:SpecifiedLineTradeSettlement")
    tax = ET.SubElement(settlement, "ram:ApplicableTradeTax")
    ET.SubElement(tax, "ram:TypeCode").text = "VAT"
    ET.SubElement(tax, "ram:CategoryCode").text = line_item.tax_category
    if line_item.tax_rate is not None:
        ET.SubElement(tax, "ram:RateApplicablePercent").text = str(
            line_item.tax_rate
        )
    if isinstance(line_item, EN16931LineItem):
        if line_item.billing_period is not None:
            start, end = line_item.billing_period
            assert start <= end
            period_el = ET.SubElement(settlement, "ram:BillingSpecifiedPeriod")
            _date_element(period_el, "ram:StartDateTime", start)
            _date_element(period_el, "ram:EndDateTime", end)
    for allowance in line_item.allowances:
        _generate_allowance_or_charge(
            settlement,
            "ram:SpecifiedTradeAllowanceCharge",
            invoice,
            allowance,
            False,
        )
    for charge in line_item.charges:
        _generate_allowance_or_charge(
            settlement,
            "ram:SpecifiedTradeAllowanceCharge",
            invoice,
            charge,
            True,
        )
    summation = ET.SubElement(
        settlement,
        "ram:SpecifiedTradeSettlementLineMonetarySummation",
    )
    _currency_element(
        summation,
        "ram:LineTotalAmount",
        line_item.billed_total,
        invoice.currency_code,
    )
    if isinstance(line_item, EN16931LineItem):
        if line_item.doc_ref is not None:
            _generate_referenced_document(settlement, line_item.doc_ref)
        if line_item.trade_account_id is not None:
            _generate_trade_account(settlement, line_item.trade_account_id)


def _generate_preceding_invoice(
    parent: ET.Element, doc: tuple[str, datetime.date | None]
) -> None:
    id, date = doc
    doc_el = ET.SubElement(parent, "ram:InvoiceReferencedDocument")
    ET.SubElement(doc_el, "ram:IssuerAssignedID").text = id
    if date is not None:
        _date_element(doc_el, "ram:FormattedIssueDateTime", date)


def _generate_referenced_document(parent: ET.Element, doc: DocRef) -> None:
    id, ref_type_code = doc
    doc_el = ET.SubElement(parent, "ram:AdditionalReferencedDocument")
    if id is not None:
        ET.SubElement(doc_el, "ram:IssuerAssignedID").text = id
    ET.SubElement(doc_el, "ram:TypeCode").text = str(
        DocumentTypeCode.INVOICING_DATA_SHEET
    )
    if ref_type_code is not None:
        ET.SubElement(doc_el, "ram:ReferenceTypeCode").text = ref_type_code


def _generate_trade_account(parent: ET.Element, id: str) -> None:
    _id_element(parent, "ram:ReceivableSpecifiedTradeAccountingAccount", id)


def _generate_allowance_or_charge(
    parent: ET.Element,
    name: str,
    invoice: BasicWLInvoice,
    allowance_or_charge: LineAllowance | LineCharge,
    is_charge: bool,
) -> None:
    el = ET.SubElement(parent, name)
    indicator_el = ET.SubElement(el, "ram:ChargeIndicator")
    ET.SubElement(indicator_el, "udt:Indicator").text = (
        "true" if is_charge else "false"
    )
    if allowance_or_charge.percent is not None:
        ET.SubElement(el, "ram:CalculationPercent").text = str(
            allowance_or_charge.percent
        )
    if allowance_or_charge.basis_amount is not None:
        _currency_element(
            el,
            "ram:BasisAmount",
            allowance_or_charge.basis_amount,
            invoice.currency_code,
        )
    _currency_element(
        el,
        "ram:ActualAmount",
        allowance_or_charge.actual_amount,
        invoice.currency_code,
    )
    if allowance_or_charge.reason_code is not None:
        ET.SubElement(el, "ram:ReasonCode").text = str(
            allowance_or_charge.reason_code
        )
    if allowance_or_charge.reason is not None:
        ET.SubElement(el, "ram:Reason").text = allowance_or_charge.reason
    if isinstance(allowance_or_charge, (DocumentAllowance, DocumentCharge)):
        tax_el = ET.SubElement(el, "ram:CategoryTradeTax")
        ET.SubElement(tax_el, "ram:TypeCode").text = "VAT"
        ET.SubElement(tax_el, "ram:CategoryCode").text = str(
            allowance_or_charge.tax_category
        )
        if allowance_or_charge.tax_rate is not None:
            ET.SubElement(tax_el, "ram:RateApplicablePercent").text = str(
                allowance_or_charge.tax_rate
            )


def _generate_trade_agreement(
    parent: ET.Element, invoice: MinimumInvoice
) -> None:
    agreement_el = ET.SubElement(parent, "ram:ApplicableHeaderTradeAgreement")
    if invoice.buyer_reference is not None:
        ET.SubElement(
            agreement_el, "ram:BuyerReference"
        ).text = invoice.buyer_reference
    _generate_trade_party(agreement_el, "ram:SellerTradeParty", invoice.seller)
    _generate_trade_party(agreement_el, "ram:BuyerTradeParty", invoice.buyer)
    if (
        isinstance(invoice, BasicWLInvoice)
        and invoice.seller_tax_representative is not None
    ):
        _generate_trade_party(
            agreement_el,
            "ram:SellerTaxRepresentativeTradeParty",
            invoice.seller_tax_representative,
        )
    if isinstance(invoice, EN16931Invoice):
        _document_element(
            agreement_el,
            "ram:SellerOrderReferencedDocument",
            invoice.seller_order_id,
        )
    _document_element(
        agreement_el,
        "ram:BuyerOrderReferencedDocument",
        invoice.buyer_order_id,
    )
    if isinstance(invoice, BasicWLInvoice):
        _document_element(
            agreement_el,
            "ram:ContractReferencedDocument",
            invoice.contract_id,
        )
    if isinstance(invoice, EN16931Invoice):
        for doc in invoice.referenced_docs:
            _generate_referenced_doc(agreement_el, doc)
        if invoice.procuring_project is not None:
            project_id, project_name = invoice.procuring_project
            project_el = ET.SubElement(
                agreement_el, "ram:SpecifiedProcuringProject"
            )
            ET.SubElement(project_el, "ram:ID").text = project_id
            ET.SubElement(project_el, "ram:Name").text = project_name


def _generate_referenced_doc(
    parent: ET.Element, doc: ReferenceDocument
) -> None:
    el = ET.SubElement(parent, "ram:AdditionalReferencedDocument")
    ET.SubElement(el, "ram:IssuerAssignedID").text = doc.id
    if doc.url is not None:
        ET.SubElement(el, "ram:URIID").text = doc.url
    ET.SubElement(el, "ram:TypeCode").text = str(doc.type_code)
    if doc.name is not None:
        ET.SubElement(el, "ram:Name").text = doc.name
    if doc.attachment is not None:
        content, mime_type, filename = doc.attachment
        attach_el = ET.SubElement(el, "ram:AttachmentBinaryObject")
        attach_el.text = b64encode(content).decode("ascii")
        attach_el.set("mimeCode", mime_type)
        attach_el.set("filename", filename)
    if doc.reference_type_code is not None:
        ET.SubElement(
            el, "ram:ReferenceTypeCode"
        ).text = doc.reference_type_code


def _generate_delivery(parent: ET.Element, invoice: MinimumInvoice) -> None:
    delivery_el = ET.SubElement(parent, "ram:ApplicableHeaderTradeDelivery")

    if isinstance(invoice, BasicWLInvoice):
        if invoice.ship_to is not None:
            _generate_trade_party(
                delivery_el, "ram:ShipToTradeParty", invoice.ship_to
            )
        if invoice.delivery_date is not None:
            supply_el = ET.SubElement(
                delivery_el, "ram:ActualDeliverySupplyChainEvent"
            )
            _date_element(
                supply_el, "ram:OccurrenceDateTime", invoice.delivery_date
            )
        if invoice.despatch_advice_id is not None:
            _document_element(
                delivery_el,
                "ram:DespatchAdviceReferencedDocument",
                invoice.despatch_advice_id,
            )
    if isinstance(invoice, EN16931Invoice):
        if invoice.receiving_advice_id is not None:
            _document_element(
                delivery_el,
                "ram:ReceivingAdviceReferencedDocument",
                invoice.receiving_advice_id,
            )


def _generate_settlement(parent: ET.Element, invoice: MinimumInvoice) -> None:
    settlement_el = ET.SubElement(
        parent, "ram:ApplicableHeaderTradeSettlement"
    )
    if isinstance(invoice, BasicWLInvoice):
        if invoice.sepa_reference is not None:
            ET.SubElement(
                settlement_el, "ram:CreditorReferenceID"
            ).text = invoice.sepa_reference
        if invoice.payment_reference is not None:
            ET.SubElement(
                settlement_el, "ram:PaymentReference"
            ).text = invoice.payment_reference
    if isinstance(invoice, EN16931Invoice):
        if invoice.tax_currency_code is not None:
            ET.SubElement(
                settlement_el, "ram:TaxCurrencyCode"
            ).text = invoice.tax_currency_code
    ET.SubElement(
        settlement_el, "ram:InvoiceCurrencyCode"
    ).text = invoice.currency_code

    if isinstance(invoice, BasicWLInvoice):
        if invoice.payee is not None:
            _generate_trade_party(
                settlement_el, "ram:PayeeTradeParty", invoice.payee
            )
        for means in invoice.payment_means:
            _generate_payment_means(settlement_el, means)
        for tax in invoice.tax:
            _generate_tax(settlement_el, invoice, tax)
        if invoice.billing_period is not None:
            billing_period = ET.SubElement(
                settlement_el, "ram:BillingSpecifiedPeriod"
            )
            start, end = invoice.billing_period
            _date_element(billing_period, "ram:StartDateTime", start)
            _date_element(billing_period, "ram:EndDateTime", end)

        for allowance in invoice.allowances:
            _generate_allowance_or_charge(
                settlement_el,
                "ram:SpecifiedTradeAllowanceCharge",
                invoice,
                allowance,
                False,
            )
        for charge in invoice.charges:
            _generate_allowance_or_charge(
                settlement_el,
                "ram:SpecifiedTradeAllowanceCharge",
                invoice,
                charge,
                True,
            )

        if invoice.payment_terms is not None:
            _generate_payment_terms(settlement_el, invoice.payment_terms)

    _generate_summation(settlement_el, invoice)

    if isinstance(invoice, BasicWLInvoice):
        for preceding in invoice.preceding_invoices:
            _generate_preceding_invoice(settlement_el, preceding)
        for ref_id in invoice.receiver_accounting_ids:
            _generate_trade_account(settlement_el, ref_id)


def _generate_payment_means(parent: ET.Element, means: PaymentMeans) -> None:
    means_el = ET.SubElement(
        parent, "ram:SpecifiedTradeSettlementPaymentMeans"
    )
    ET.SubElement(means_el, "ram:TypeCode").text = str(means.type_code)
    if means.information is not None:
        ET.SubElement(means_el, "ram:Information").text = means.information
    if means.card is not None:
        card_id, cardholder = means.card
        card_el = ET.SubElement(
            means_el, "ram:ApplicableTradeSettlementFinancialCard"
        )
        ET.SubElement(card_el, "ram:ID").text = card_id
        if cardholder is not None:
            ET.SubElement(card_el, "ram:CardholderName").text = cardholder
    if means.payer_iban is not None:
        account_el = ET.SubElement(
            means_el, "ram:PayerPartyDebtorFinancialAccount"
        )
        ET.SubElement(account_el, "ram:IBANID").text = means.payer_iban
    if means.payee_account is not None:
        account_el = ET.SubElement(
            means_el, "ram:PayeePartyCreditorFinancialAccount"
        )
        if means.payee_account.iban is not None:
            ET.SubElement(
                account_el, "ram:IBANID"
            ).text = means.payee_account.iban
        if means.payee_account.name is not None:
            ET.SubElement(
                account_el, "ram:AccountName"
            ).text = means.payee_account.name
        if means.payee_account.bank_id is not None:
            ET.SubElement(
                account_el, "ram:ProprietaryID"
            ).text = means.payee_account.bank_id
    if means.payee_bic is not None:
        bic_el = ET.SubElement(
            means_el, "ram:PayeeSpecifiedCreditorFinancialInstitution"
        )
        ET.SubElement(bic_el, "ram:BICID").text = means.payee_bic


def _generate_tax(
    parent: ET.Element, invoice: MinimumInvoice, tax: Tax
) -> None:
    tax_el = ET.SubElement(parent, "ram:ApplicableTradeTax")
    _currency_element(
        tax_el,
        "ram:CalculatedAmount",
        tax.calculated_amount,
        invoice.currency_code,
    )
    ET.SubElement(tax_el, "ram:TypeCode").text = "VAT"
    if tax.exemption_reason is not None:
        ET.SubElement(
            tax_el, "ram:ExemptionReason"
        ).text = tax.exemption_reason
    _currency_element(
        tax_el, "ram:BasisAmount", tax.basis_amount, invoice.currency_code
    )
    ET.SubElement(tax_el, "ram:CategoryCode").text = str(tax.category_code)
    if tax.exemption_reason_code is not None:
        ET.SubElement(tax_el, "ram:ExemptionReasonCode").text = str(
            tax.exemption_reason_code
        )
    if tax.tax_point_date is not None:
        _date_element(tax_el, "ram:TaxPointDate", tax.tax_point_date)
    if tax.due_date_type_code is not None:
        ET.SubElement(tax_el, "ram:DueDateTypeCode").text = str(
            tax.due_date_type_code
        )
    if tax.rate_percent is not None:
        ET.SubElement(tax_el, "ram:RateApplicablePercent").text = str(
            tax.rate_percent
        )


def _generate_payment_terms(parent: ET.Element, terms: PaymentTerms) -> None:
    terms_el = ET.SubElement(parent, "ram:SpecifiedTradePaymentTerms")
    if terms.description is not None:
        ET.SubElement(terms_el, "ram:Description").text = terms.description
    if terms.due_date is not None:
        _date_element(terms_el, "ram:DueDateDateTime", terms.due_date)
    if terms.direct_debit_mandate_id is not None:
        ET.SubElement(
            terms_el, "ram:DirectDebitMandateID"
        ).text = terms.direct_debit_mandate_id


def _generate_summation(parent: ET.Element, invoice: MinimumInvoice) -> None:
    summation = ET.SubElement(
        parent, "ram:SpecifiedTradeSettlementHeaderMonetarySummation"
    )
    if invoice.line_total_amount is not None:
        _currency_element(
            summation,
            "ram:LineTotalAmount",
            invoice.line_total_amount,
            invoice.currency_code,
        )
    if isinstance(invoice, BasicWLInvoice):
        if invoice.charge_total_amount is not None:
            _currency_element(
                summation,
                "ram:ChargeTotalAmount",
                invoice.charge_total_amount,
                invoice.currency_code,
            )
        if invoice.allowance_total_amount is not None:
            _currency_element(
                summation,
                "ram:AllowanceTotalAmount",
                invoice.allowance_total_amount,
                invoice.currency_code,
            )
    _currency_element(
        summation,
        "ram:TaxBasisTotalAmount",
        invoice.tax_basis_total_amount,
        invoice.currency_code,
    )
    for amount in invoice.tax_total_amounts:
        _currency_element(
            summation,
            "ram:TaxTotalAmount",
            amount,
            invoice.currency_code,
            with_currency=True,
        )
    if isinstance(invoice, EN16931Invoice):
        if invoice.rounding_amount is not None:
            _currency_element(
                summation,
                "ram:RoundingAmount",
                invoice.rounding_amount,
                invoice.currency_code,
            )
    _currency_element(
        summation,
        "ram:GrandTotalAmount",
        invoice.grand_total_amount,
        invoice.currency_code,
    )
    if isinstance(invoice, BasicWLInvoice):
        if invoice.prepaid_amount is not None:
            _currency_element(
                summation,
                "ram:TotalPrepaidAmount",
                invoice.prepaid_amount,
                invoice.currency_code,
            )
    _currency_element(
        summation,
        "ram:DuePayableAmount",
        invoice.due_payable_amount,
        invoice.currency_code,
    )


if __name__ == "__main__":
    from ._test_data import TEST_EN16931_INVOICE

    tree = generate_et(TEST_EN16931_INVOICE)
    ET.indent(tree)
    print(ET.tostring(tree, encoding="unicode", xml_declaration=True))



import sys
from pathlib import Path

from pypdf import PdfReader
from pypdf.errors import PdfReadError

from ._locale import setup_locale
from .exc import NoFacturXError, PDFParseError
from .pdf_common import FACTURX_FILENAME, FileRelationship

_ = setup_locale()


def extract_facturx_from_pdf(
    filename: str | Path,
) -> tuple[str, FileRelationship | None]:
    """Extract the Factur-X XML file from a PDF file.

    If the PDF file cannot be processed, a PDFParseError is raised. If it
    does not contain a Factur-X XML file, a NoFacturXError is raised.
    """

    try:
        pdf = PdfReader(filename)
    except PdfReadError as exc:
        raise PDFParseError(_("Cannot read PDF file: {}").format(exc)) from exc
    try:
        # TODO: Support /Kids nodes
        doc = pdf.trailer["/Root"]["/Names"]["/EmbeddedFiles"]["/Names"]  # type: ignore[index]
        while doc:
            if doc[0] == FACTURX_FILENAME:
                break
        obj = doc[1]
        relationship: FileRelationship | None = None
        rel_s = obj.get("/AFRelationship")
        if rel_s is not None:
            relationship = FileRelationship(obj["/AFRelationship"][1:])
        file = obj["/EF"]["/F"]
        if file["/Subtype"] != "/text/xml":
            raise NoFacturXError(_("No Factur-X invoice found in PDF file"))
    except (KeyError, IndexError, ValueError) as exc:
        raise NoFacturXError(
            _("No Factur-X invoice found in PDF file")
        ) from exc
    return file.get_data().decode("utf-8"), relationship


def main() -> None:
    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} PDF-FILE", file=sys.stderr)
        sys.exit(1)
    stream, relationship = extract_facturx_from_pdf(sys.argv[1])
    print("Relationship:", relationship)
    with open("facturx.xml", "w") as target_stream:
        target_stream.write(stream)


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import Sequence
from gettext import ngettext
from textwrap import indent

from ._locale import setup_locale
from .model import (
    BankAccount,
    BasicInvoice,
    BasicWLInvoice,
    DocumentAllowance,
    DocumentCharge,
    EN16931Invoice,
    IncludedNote,
    LineAllowance,
    LineCharge,
    LineItem,
    MinimumInvoice,
    PaymentMeans,
    PaymentTerms,
    PostalAddress,
    ReferenceDocument,
    TradeContact,
    TradeParty,
)
from .quantities import QUANTITY_NAMES
from .type_codes import (
    DOCUMENT_TYPE_NAMES,
    PAYMENT_MEANS_NAMES,
    PAYMENT_TIME_CODE_NAMES,
    REFERENCE_QUALIFIER_NAMES,
    TEXT_SUBJECT_CODE_NAMES,
    IdentifierSchemeCode,
)
from .types import ID, OptionalQuantity, Quantity

__all__ = ["format_invoice_as_text", "format_trade_party", "format_address"]

_ = setup_locale()


def format_invoice_as_text(invoice: MinimumInvoice) -> str:
    lines: list[str] = []

    type_name = _(DOCUMENT_TYPE_NAMES[invoice.type_code])
    lines += [
        (
            "{document_type} {invoice_number}".format(
                document_type=type_name, invoice_number=invoice.invoice_number
            )
        ),
        _("Date: {invoice_date:%Y-%m-%d}").format(
            invoice_date=invoice.invoice_date
        ),
    ]
    if isinstance(invoice, BasicWLInvoice):
        if invoice.delivery_date:
            lines.append(
                _("Delivery Date: {delivery_date:%Y-%m-%d}").format(
                    delivery_date=invoice.delivery_date
                )
            )
        if invoice.billing_period:
            start, end = invoice.billing_period
            lines.append(
                _("Billing Period: {start:%Y-%m-%d}–{end:%Y-%m-%d}").format(
                    start=start, end=end
                )
            )

    lines += [
        _header(_("Sender")),
        format_trade_party(invoice.seller),
        _header(_("Recipient")),
        format_trade_party(invoice.buyer),
    ]
    if isinstance(invoice, BasicWLInvoice):
        if invoice.ship_to:
            lines += [
                _header(_("Ship To")),
                format_trade_party(invoice.ship_to),
            ]
        if invoice.seller_tax_representative:
            lines += [
                _header(_("Seller's Tax Representative")),
                format_trade_party(invoice.seller_tax_representative),
            ]

    refs = _format_references(invoice)
    if refs:
        lines += [_header(_("References")), refs]
    if isinstance(invoice, BasicWLInvoice):
        if invoice.notes:
            formatted_notes = [format_note(note) for note in invoice.notes]
            lines += [
                _header(_("Notes")),
                "\n\n----\n\n".join(formatted_notes),
            ]
    if isinstance(invoice, BasicInvoice):
        lines += [
            _header(_("Line Items")),
            _format_line_items(invoice.line_items),
        ]
    if isinstance(invoice, BasicWLInvoice):
        if invoice.charges:
            lines += [
                _header(_("Surcharges")),
                _format_allowances_and_charges(invoice.charges),
            ]
        if invoice.allowances:
            lines += [
                _header(_("Deductions")),
                _format_allowances_and_charges(invoice.allowances),
            ]
        lines += [
            _header(_("Tax")),
            _format_tax(invoice),
        ]
    lines += [
        _header(_("Totals")),
        _format_totals(invoice),
    ]
    if isinstance(invoice, BasicWLInvoice):
        lines.append(_format_payment(invoice))

    return "\n".join(lines)


def _format_references(invoice: MinimumInvoice) -> str:
    lines = []
    if isinstance(invoice, BasicWLInvoice):
        for preceding in invoice.preceding_invoices:
            number, date = preceding
            if date:
                lines.append(
                    _("Related Invoice: {number} ({date:%Y-%m-%d})").format(
                        number=number, date=date
                    )
                )
            else:
                lines.append(
                    _("Related Invoice: {number}").format(number=number)
                )
    if isinstance(invoice, BasicWLInvoice) and invoice.contract_id:
        lines.append(_("Contract ID: {}").format(invoice.contract_id))
    if isinstance(invoice, EN16931Invoice) and invoice.procuring_project:
        id, name = invoice.procuring_project
        lines.append(
            _("Procuring Project: {id} {name}").format(id=id, name=name)
        )
    if invoice.business_process_id is not None:
        lines.append(
            _("Business Process ID: {}").format(invoice.business_process_id)
        )
    if invoice.buyer_reference is not None:
        lines.append(_("Buyer Reference: {}").format(invoice.buyer_reference))
    if invoice.buyer_order_id is not None:
        lines.append(_("Buyer Order ID: {}").format(invoice.buyer_order_id))
    if isinstance(invoice, EN16931Invoice) and invoice.seller_order_id:
        lines.append(_("Seller Order ID: {}").format(invoice.seller_order_id))
    if isinstance(invoice, BasicWLInvoice) and invoice.despatch_advice_id:
        lines.append(
            _("Despatch Advice ID: {}").format(invoice.despatch_advice_id)
        )
    if isinstance(invoice, EN16931Invoice) and invoice.receiving_advice_id:
        lines.append(
            _("Receiving Advice ID: {}").format(invoice.receiving_advice_id)
        )
    if isinstance(invoice, BasicWLInvoice) and invoice.despatch_advice_id:
        for id in invoice.receiver_accounting_ids:
            lines.append(_("Receiver Accounting ID: {}").format(id))
    if isinstance(invoice, EN16931Invoice):
        for doc in invoice.referenced_docs:
            lines.append(format_reference_doc(doc))
    return "\n".join(lines)


def _format_line_items(line_items: Sequence[LineItem]) -> str:
    lines = []
    for li in line_items:
        lines.append(
            _("{id} {net_price} {quantity} {total_price}").format(
                id=li.id,
                net_price=li.net_price,
                quantity=format_quantity(li.billed_quantity),
                total_price=li.billed_total,
            )
        )
        for allowance in li.allowances:
            lines.append(_format_line_allowance_or_charge(allowance))
        for charge in li.charges:
            lines.append(_format_line_allowance_or_charge(charge))
        if li.basis_quantity:
            lines.append(
                _("  Basis Quantity: {}").format(
                    format_quantity(li.basis_quantity)
                )
            )
        if li.tax_rate is not None:
            lines.append(
                _("  VAT: {tax_rate}\u2009% ({tax_category})").format(
                    tax_rate=li.tax_rate, tax_category=li.tax_category
                )
            )
        lines.append(indent(li.name, "  "))
        if li.global_id:
            lines.append(
                "  {global_id}".format(
                    global_id=_format_global_id(li.global_id)
                )
            )

    return "\n".join(lines)


def _format_line_allowance_or_charge(
    allowance: LineAllowance | LineCharge,
) -> str:
    lines = []
    type = (
        _("Surcharge") if isinstance(allowance, LineCharge) else _("Deduction")
    )
    lines.append(
        _("  {type}: {basis} {percent:>4}\u2009% {amount}").format(
            type=type,
            basis=allowance.basis_amount,
            percent=allowance.percent or "",
            amount=allowance.actual_amount or "",
        )
    )
    if allowance.reason and allowance.reason_code:
        lines.append(
            _("  Reason: {reason} ({code})").format(
                reason=allowance.reason, code=allowance.reason_code.name
            )
        )
    elif allowance.reason:
        lines.append(_("  Reason: {reason}").format(reason=allowance.reason))
    elif allowance.reason_code:
        lines.append(
            _("  Reason: {code}").format(code=allowance.reason_code.name)
        )
    return "\n".join(lines)


def _format_allowances_and_charges(
    allowances: Sequence[DocumentAllowance | DocumentCharge],
) -> str:
    basis_len = max(
        (
            len(str(a.basis_amount))
            for a in allowances
            if a.basis_amount is not None
        ),
        default=0,
    )
    amount_len = max(
        (
            len(str(a.actual_amount))
            for a in allowances
            if a.actual_amount is not None
        ),
        default=0,
    )
    basis_s = _("Basis").center(basis_len)
    amount_s = _("Amount").center(amount_len)
    lines = [
        _("{basis} |   %    |   Tax    | {amount}").format(
            basis=basis_s, amount=amount_s
        )
    ]
    for a in allowances:
        basis_s = str(a.basis_amount).rjust(basis_len)
        amount_s = str(a.actual_amount).rjust(amount_len)
        lines.append(
            _(
                "{basis} | {percent:>4}\u2009% |"
                " {tax_rate}\u2009% ({tax_category}) | {amount}"
            ).format(
                basis=basis_s,
                percent=a.percent,
                tax_rate=a.tax_rate,
                tax_category=a.tax_category.name,
                amount=amount_s,
            )
        )
        if a.reason and a.reason_code:
            lines.append(
                _("  Reason: {reason} ({code})").format(
                    reason=a.reason, code=a.reason_code.name
                )
            )
        elif a.reason:
            lines.append(_("  Reason: {reason}").format(reason=a.reason))
        elif a.reason_code:
            lines.append(_("  Reason: {code}").format(code=a.reason_code.name))
    return "\n".join(lines)


def _format_tax(invoice: BasicWLInvoice) -> str:
    lines = []

    if isinstance(invoice, EN16931Invoice) and invoice.tax_currency_code:
        lines += [_("Tax Currency: {}").format(invoice.tax_currency_code), ""]

    basis_len = max(
        (len(str(tax.basis_amount)) for tax in invoice.tax), default=0
    )
    tax_len = max(
        (len(str(tax.calculated_amount)) for tax in invoice.tax), default=0
    )
    net_s = _("Net").center(basis_len)
    tax_s = _("Tax").center(tax_len)
    lines.append(
        _("Cat | Rate | {net} | {tax}").format(net=net_s, tax=tax_s),
    )
    for tax in invoice.tax:
        basis_s = str(tax.basis_amount).rjust(basis_len)
        tax_s = str(tax.calculated_amount).rjust(tax_len)
        lines.append(
            _("{category:>2}  | {rate:>2}\u2009% | {basis} | {tax}").format(
                category=tax.category_code,
                rate=tax.rate_percent,
                basis=basis_s,
                tax=tax_s,
            )
        )

        exemption = ""
        if tax.exemption_reason and tax.exemption_reason_code:
            exemption += _("  Exemption Reason: {reason} ({code})").format(
                reason=tax.exemption_reason,
                code=tax.exemption_reason_code.name,
            )
        elif tax.exemption_reason:
            exemption += _("  Exemption Reason: {reason}").format(
                reason=tax.exemption_reason
            )
        elif tax.exemption_reason_code:
            exemption += _("  Exemption Reason: {code}").format(
                code=tax.exemption_reason_code.name
            )
        if exemption:
            lines.append(exemption)

        due_date = ""
        if tax.tax_point_date and tax.due_date_type_code:
            time_code = _(PAYMENT_TIME_CODE_NAMES[tax.due_date_type_code])
            due_date += _(
                "  Tax Point Date: {date:%Y-%m-%d} ({time_code})"
            ).format(
                date=tax.tax_point_date,
                time_code=time_code,
            )
        elif tax.tax_point_date:
            due_date += _("  Tax Point Date: {date:%Y-%m-%d}").format(
                date=tax.tax_point_date
            )
        elif tax.due_date_type_code:
            time_code = _(PAYMENT_TIME_CODE_NAMES[tax.due_date_type_code])
            due_date += _("  Due Date Type: {time_code}").format(
                time_code=time_code
            )
        if due_date:
            lines.append(due_date)

    return "\n".join(lines)


def _format_totals(invoice: MinimumInvoice) -> str:
    lines = []
    if invoice.line_total_amount:
        lines.append(
            _("Line Total (net): {}").format(invoice.line_total_amount)
        )
    if isinstance(invoice, BasicWLInvoice):
        if invoice.charge_total_amount:
            lines.append(
                _("Surcharges: {}").format(invoice.charge_total_amount)
            )
        if invoice.allowance_total_amount:
            lines.append(
                _("Deductions: {}").format(invoice.allowance_total_amount)
            )
    lines.append(_("Net: {}").format(invoice.tax_basis_total_amount))
    for tax_amount in invoice.tax_total_amounts:
        lines.append(_("VAT: {}").format(tax_amount))
    lines.append(_("Gross: {}").format(invoice.grand_total_amount))
    if isinstance(invoice, BasicWLInvoice) and invoice.prepaid_amount:
        lines.append(_("Prepaid: {}").format(invoice.prepaid_amount))
    if isinstance(invoice, EN16931Invoice) and invoice.rounding_amount:
        lines.append(_("Rounding amount: {}").format(invoice.rounding_amount))
    lines.append(_("Amount payable: {}").format(invoice.due_payable_amount))
    return "\n".join(lines)


def _format_payment(invoice: BasicWLInvoice) -> str:
    lines = []
    if invoice.payee:
        lines += [
            _header(_("Payee")),
            format_trade_party(invoice.payee),
        ]
    if (
        invoice.payment_means
        or invoice.sepa_reference
        or invoice.payment_reference
    ):
        lines.append(_header(_("Payment Means")))
        for means in invoice.payment_means:
            lines.append(format_payment_means(means))
        if invoice.sepa_reference:
            lines.append(
                _("SEPA Reference: {}").format(invoice.sepa_reference)
            )
        if invoice.payment_reference:
            lines.append(
                _("Payment Reference: {}").format(invoice.payment_reference)
            )
    if invoice.payment_terms:
        lines += [
            _header(_("Payment Terms")),
            format_payment_terms(invoice.payment_terms),
        ]
    return "\n".join(lines)


def format_quantity(quantity: Quantity | OptionalQuantity) -> str:
    if quantity[1] is None:
        return str(quantity[0])

    qn = QUANTITY_NAMES[quantity[1]]
    if isinstance(qn, tuple):
        q = quantity[0].as_integer_ratio()
        if q[1] == 1:  # if unit is an integer
            unit = ngettext(*qn, q[0])
        else:
            unit = _(qn[1])  # plural
    else:
        unit = qn
    return _("{quantity} {unit}").format(quantity=quantity[0], unit=unit)


def format_trade_party(trade_party: TradeParty) -> str:
    lines = []

    if trade_party.name:
        if trade_party.trading_business_name:
            lines.append(
                _("{name} ({trading_business_name})").format(
                    name=trade_party.name,
                    trading_business_name=trade_party.trading_business_name,
                )
            )
        else:
            lines.append(trade_party.name)
    if trade_party.description:
        lines.append(trade_party.description)
    if trade_party.address:
        lines.append(format_address(trade_party.address))
    if trade_party.email:
        lines.append(trade_party.email)
    if trade_party.vat_id:
        lines.append(_("VAT ID: {}").format(trade_party.vat_id))
    if trade_party.tax_number:
        lines.append(_("Tax Number: {}").format(trade_party.tax_number))
    if len(trade_party.ids) == 1:
        lines.append(_("ID: {}").format(trade_party.ids[0]))
    elif len(trade_party.ids) > 1:
        lines.append(_("IDs: {}").format(", ".join(trade_party.ids)))
    for gid in trade_party.global_ids:
        lines.append(_format_global_id(gid))
    if trade_party.legal_id:
        lines.append(_format_global_id(trade_party.legal_id))
    if trade_party.contact is not None:
        lines.append(format_trade_contact(trade_party.contact))

    return "\n".join(lines)


def format_address(address: PostalAddress) -> str:
    lines = []
    if address.line_one:
        lines.append(address.line_one)
    if address.line_two:
        lines.append(address.line_two)
    if address.line_three:
        lines.append(address.line_three)
    city_line = address.country_code
    if address.post_code:
        city_line += f" {address.post_code}"
    if address.city:
        city_line += f" {address.city}"
    lines.append(city_line)
    if address.country_subdivision:
        lines.append(address.country_subdivision)
    return "\n".join(lines)


def format_trade_contact(contact: TradeContact) -> str:
    lines = [_("Contact:")]
    if contact.person_name:
        lines.append(f"  {contact.person_name}")
    elif contact.department_name:
        lines.append(f"  {contact.department_name}")
    if contact.phone:
        lines.append(_("  Phone: {phone}").format(phone=contact.phone))
    if contact.email:
        lines.append(f"  {contact.email}")
    return "\n".join(lines)


def format_note(note: IncludedNote) -> str:
    if note.subject_code:
        code = _(TEXT_SUBJECT_CODE_NAMES[note.subject_code])
        return _("{subject_code}: {content}").format(
            subject_code=code, content=note.content.strip()
        )
    else:
        return note.content.strip()


def format_payment_means(payment_means: PaymentMeans) -> str:
    lines = []
    lines.append(
        _("Payment Means: {}").format(
            _(PAYMENT_MEANS_NAMES[payment_means.type_code])
        )
    )
    if payment_means.payee_account:
        lines.append(format_bank_account(payment_means.payee_account))
    if payment_means.payee_bic:
        lines.append(_("BIC: {}").format(payment_means.payee_bic))
    if payment_means.card:
        pan, cardholder = payment_means.card
        if cardholder:
            lines.append(_("Credit card: {} ({})").format(pan, cardholder))
        else:
            lines.append(_("Credit card: {}").format(pan))
    if payment_means.payer_iban:
        lines.append(_("Payer IBAN: {}").format(payment_means.payer_iban))
    if payment_means.information:
        lines.append(payment_means.information)
    return "\n".join(lines)


def format_bank_account(account: BankAccount) -> str:
    lines = []
    if account.iban:
        lines.append(_("IBAN: {}").format(account.iban))
    if account.name:
        lines.append(_("Account owner: {}").format(account.name))
    if account.bank_id:
        lines.append(_("Bank: {}").format(account.bank_id))
    return "\n".join(lines)


def format_payment_terms(terms: PaymentTerms) -> str:
    lines = []
    if terms.due_date:
        lines.append(
            _("Due Date: {due_date:%Y-%m-%d}").format(due_date=terms.due_date)
        )
    if terms.direct_debit_mandate_id:
        lines.append(
            _("Direct Debit Mandate ID: {}").format(
                terms.direct_debit_mandate_id
            )
        )
    description = terms.description.strip() if terms.description else ""
    if description:
        lines.append(description)
    return "\n".join(lines)


def format_reference_doc(doc: ReferenceDocument) -> str:
    lines = []
    type_name = _(DOCUMENT_TYPE_NAMES[doc.type_code])
    lines.append("{type} {doc.id}".format(type=type_name, doc=doc))
    if doc.reference_type_code is not None:
        lines.append(
            _("  Reference Qualifier: {}").format(
                REFERENCE_QUALIFIER_NAMES[doc.reference_type_code]
            )
        )
    if doc.name is not None:
        lines.append(_("  Name: {}").format(doc.name))
    if doc.url is not None:
        lines.append(_("  URL: {}").format(doc.url))
    if doc.attachment is not None:
        __, mime_type, filename = doc.attachment
        lines.append(
            _("  Attachment: {filename} ({mime_type})").format(
                filename=filename, mime_type=mime_type
            )
        )

    return "\n".join(lines)


def _format_global_id(global_id: ID) -> str:
    id, code = global_id
    if code is None:
        return _("Global ID: {}").format(id)
    if isinstance(code, IdentifierSchemeCode):
        code = code.name
    return _("{id_type}: {id}").format(id_type=code, id=id)


def _header(header: str) -> str:
    return f"\n{header}\n{'-' * len(header)}\n"


if __name__ == "__main__":
    # Print a sample invoice

    import locale

    from .test_data import en16931_einfach

    locale.setlocale(locale.LC_ALL, "")

    invoice = en16931_einfach()
    print(format_invoice_as_text(invoice))
